using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;

namespace I2CDotNet
{
    
    class I2C_CmT35 : I2CDriverInterface
    {
        private IntPtr ipI2Chandle = IntPtr.Zero;
        private uint offset = 0;
        private UInt32 mode = 1;
        
        // Device CreateFile specific constants
        internal const Int32 INVALID_HANDLE_VALUE = -1;
        internal const UInt32 OPEN_EXISTING = 3;
        internal const UInt32 GENERIC_READ = 0x80000000;
        internal const UInt32 GENERIC_WRITE = 0x40000000;

 
        public I2C_CmT35(int port)
        {
            string driverName = "I2C" + port + ":";
            //First, we create a handle to the I2C proxy driver. I2C3 in our case.
            ipI2Chandle = NativeMethods.CreateFile(driverName, GENERIC_WRITE | GENERIC_READ, 0, IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
            if (ipI2Chandle.ToInt32() == INVALID_HANDLE_VALUE)
                throw new InvalidOperationException("Can't open " + driverName);
        }
       
        ~I2C_CmT35()
        {
            NativeMethods.CloseHandle(ipI2Chandle);
        }

        private void CalculateMode()
        {
            if (offset >= 0x100)
            {
                //  Special values for address mode setting:
                //
                //      0x1xx   Mode 0  Offset = xx
                //      0x2xx   Mode 8  Offset = xx
                //      0x3xx   Mode 16 Offset = xx
                //      0x4xx   Mode 24 Offset = xx
                //      0x5xx   Mode 32 Offset = xx

                if (offset > 0x504)
                    throw new InvalidOperationException("Read can't cross page boundary!");
                mode = (offset >> 8) - 1;
                offset = offset & 0x000000FF;
            }
        }

        private bool SendIOCTL(uint ioctl, uint value) 
        {
            uint bytesReturned = 0;
            // Initialize unmanged memory to hold the struct.
            IntPtr inputParametersPtr = IntPtr.Zero;
            IntPtr outputParametersPtr = IntPtr.Zero;
            uint inputParametersSize = 0;
            uint outputParametersSize = 0;
            try
            {
                inputParametersSize = sizeof(uint);
                inputParametersPtr = Marshal.AllocHGlobal((int)inputParametersSize);
                // Copy the struct to unmanaged memory.
                Marshal.WriteInt32(inputParametersPtr, (int)value);
                bool res = NativeMethods.DeviceIoControl(ipI2Chandle,ioctl,
                                           inputParametersPtr, inputParametersSize,
                                           outputParametersPtr, outputParametersSize,
                                           ref bytesReturned, IntPtr.Zero);
                Marshal.FreeHGlobal(inputParametersPtr);
                if (!res)
                {
                    // Throw Win32 exception with error code
                    int err = Marshal.GetLastWin32Error();
                    throw new System.ComponentModel.Win32Exception(err);
                }
            }
            catch (Exception ex)
            {
                // throw all other exceptions to upper levels
                throw ex;
            }
            return true;
        }

        private bool SetMode(uint uiAddress)
        {
            uint IOCTL_I2C_SET_SUBADDRESS_MODE = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0201);
            offset = uiAddress;
            CalculateMode();
            return SendIOCTL(IOCTL_I2C_SET_SUBADDRESS_MODE, mode);
        }

        public bool SetSlaveAddress(uint address)
        {
            uint IOCTL_I2C_SET_SLAVE_ADDRESS = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0200);
            return SendIOCTL(IOCTL_I2C_SET_SLAVE_ADDRESS, address);
        }
        
        private bool SetPointer()
        {
            NativeMethods.SetFilePointerCE(ipI2Chandle, (int)offset, ((int)IntPtr.Zero), MoveMethod.FileBeginning);
            return true;
        }


         /// <summary>
        /// Read from I2C device
        /// </summary>
        /// <param name="subAddress">Device internal address</param>
        /// <param name="buffer"> [out] The data will be written to this array</param>
        /// <param name="size"> number of bytes to read </param>
        /// <returns></returns>
        public bool Read(uint subAddress, byte[] buffer, uint size)
        {
            int bytesRead = 0;
            SetMode(subAddress);
            SetPointer();
            try
            {
                int res = NativeMethods.ReadFileCE(ipI2Chandle, buffer, (int)size, ref bytesRead, IntPtr.Zero);
                if ((res==0) || (bytesRead != size))
                {
                    // Throw Win32 exception with error code
                    int err = Marshal.GetLastWin32Error();
                    throw new System.ComponentModel.Win32Exception(err);
                }
            }
            catch (Exception ex)
            {
                // throw all other exceptions to upper levels
                throw ex;
            }
            return true;
        }

        /// <summary>
        /// Write to I2C device
        /// </summary>
        /// <param name="subAddress"> Device internal address</param>
        /// <param name="buffer"> data to be written to the device </param>
        /// <param name="size"> number of bytes to write</param>
        /// <returns></returns>
        public bool Write(uint subAddress, byte[] buffer, uint size)
        {
            int bytesWritten = 0;
            SetMode(subAddress);
            SetPointer();
            try
            {
                int res = NativeMethods.WriteFileCE(ipI2Chandle, buffer, (int)size, ref bytesWritten, IntPtr.Zero);
                if ((res == 0) || (bytesWritten != size))
                {
                    // Throw Win32 exception with error code
                    int err = Marshal.GetLastWin32Error();
                    throw new System.ComponentModel.Win32Exception(err);
                }
            }
            catch (Exception ex)
            {
                // throw all other exceptions to upper levels
                throw ex;
            }
            return true;
        }

       
    }
}
